//   Copyright (c)  2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RandomSource.H"
#include "CoCoA/error.H"

#include <cstddef>
using std::size_t;
#include <iostream>
//using std::ostream;
#include <limits>
using std::numeric_limits;


namespace CoCoA
{

  RandomSource::RandomSource(const MachineInteger& seed):
      mySeed((IsZero(seed)?time(0):abs(seed)))
  {
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
  }


  RandomSource::RandomSource(const RandomSource& copy):
      mySeed(copy.mySeed)
  {
    gmp_randinit_set(myState, copy.myState);
  }


  RandomSource& RandomSource::operator=(const RandomSource& rhs)
  {
    gmp_randclear(myState);                 // Together these effect an assignment
    gmp_randinit_set(myState, rhs.myState); //
    mySeed = rhs.mySeed;
    return *this;
  }


  RandomSource::~RandomSource()
  {
    gmp_randclear(myState);
  }


  bool RandomBool(RandomSource& src)
  {
    return gmp_urandomb_ui(src.myState, 1) == 1;
  }


  long RandomLong(RandomSource& src)
  {
    const unsigned long rnd = gmp_urandomb_ui(src.myState, numeric_limits<unsigned long>::digits);
    return static_cast<long>(rnd);
  }


  long RandomLong(RandomSource& src, const MachineInteger& lwb, const MachineInteger& upb)
  {
    const long lo = AsSignedLong(lwb);
    const long hi = AsSignedLong(upb);
    if (lo > hi) CoCoA_ERROR(ERR::BadArg, "RandomLong: must have lwb <= upb");
    const unsigned long range = 1+static_cast<unsigned long>(hi)-static_cast<unsigned long>(lo); // any overflow is harmless
    if (range == 0) return RandomLong(src);
    return lo + gmp_urandomm_ui(src.myState, range); // any overflow is harmless
  }


  ZZ RandomZZ(RandomSource& src, const MachineInteger& lwb, const MachineInteger& upb)
  {
    return RandomZZ(src, ZZ(lwb), ZZ(upb));
  }

  ZZ RandomZZ(RandomSource& src, const MachineInteger& lwb, const ZZ& upb)
  {
    return RandomZZ(src, ZZ(lwb), upb);
  }

  ZZ RandomZZ(RandomSource& src, const ZZ& lwb, const MachineInteger& upb)
  {
    return RandomZZ(src, lwb, ZZ(upb));
  }

  ZZ RandomZZ(RandomSource& src, const ZZ& lwb, const ZZ& upb)
  {
    if (lwb > upb) CoCoA_ERROR(ERR::BadArg, "RandomZZ: must have lwb <= upb");
    ZZ sample;
    mpz_urandomm(mpzref(sample), src.myState, mpzref(upb-lwb+1));
    return lwb + sample;
  }


  std::ostream& operator<<(std::ostream& out, const RandomSource& src)
  {
    out << "RandomSource(seed=" << src.mySeed << ", state=unprintable)";
    return out;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RandomSource.C,v 1.1 2010/12/17 16:03:15 abbott Exp $
// $Log: RandomSource.C,v $
// Revision 1.1  2010/12/17 16:03:15  abbott
// Added new class RandomSource.
//
//
